//--------------------------------------------------------------------
// Mandelbrot Machine (24 bits)
// by Jesús Arias (2022)
//--------------------------------------------------------------------

module SYSTEM (
	input clk,		// Main clock input 25MHz
	input reset,	// Global reset (active high)

	output xwe,		// external RAM: WE (active low)
	output xoe,		// external RAM: OE (active low)
	output [15:0]xa,// external RAM: address bus
	output [15:0]xdo,	// external RAM: output data
	input  [15:0]xdi,	// external RAM: input data

	output hsyn,	// Horizontal Sync. (active low, every 32us) 
	output vsyn,	// Vertical Sync. (active low, every 525 lines)
	output [11:0]video,	// RGB444 output

	input  [7:0]pinin,	// Input port (includes MISO)
	output [7:0]pinout,	// Output port (includes SCK, MOSI, /SS_flash, and /SS_SD)

	output	pwmout,	// Audio output (1-bit, it can be PWM modulated)

	input	rxd,	// UART
	output 	txd,
	
	output sck,		// SPI
	output mosi,
	input  miso,	
	output fss,		// Flash SS
	output sss,		// SD SS

	output	gclk,
	output	glat,
	input	gin

    
    //,output	deb0,
    //output	deb1

);

/////////////////////////////
// Pixel computing datapath
/////////////////////////////
reg [23:0]a=0;		// real part of Z
reg [23:0]b=0;		// imaginary part of Z

wire [23:0]a2;		
wire [23:0]b2;
wire [23:0]ab;
wire [24:0]abs2;

// Multipliers
fmul24 fmul0 (.md(a),.mr(a),.out(a2));
fmul24 fmul1 (.md(b),.mr(a),.out(ab));
fmul24 fmul2 (.md(b),.mr(b),.out(b2));

assign abs2=a2+b2;	// Square of the absolute value of Z
wire plus4;			// higher than 4
assign plus4=abs2[24]|abs2[23]|abs2[22]|abs2[21];

reg [8:0]icount=0;	// Iteration counter
wire init;			// Init signal
assign init=plus4 | (icount[8]);

// Sequential logic
always @(posedge clk) begin
	a<=init ? 0 : a2-b2+c;
	b<=init ? 0 : {ab[22:0],1'b0}+d;
	icount<= init ? 0 : icount+1;
end

///////////////////////////////////////////////////
// 				Image scanning
///////////////////////////////////////////////////
reg [23:0]c=146800000;	// Real part of C
reg [23:0]d=819200;		// Imaginary part of C

reg [23:0]x=(24'hffffff-96*4096+1);	// Center Position (real)
reg [23:0]y=0;				// Center Position (imaginary)
reg [12:0]delta=13'd4096;	// pixel step

reg [11:0]pix3;				// three last pixels
reg [15:0]pix4;				// buffer for 4 pixels
reg [17:0]pixcnt=0;			// Pixel counter
reg [15:0]waddr;			// write address
reg wpend=0;				// write pending

wire [3:0]pix;				// Pixel color
// Logarithmic scale for colors
assign pix= (icount[7:0]>133)?15 : (
			(icount[7:0]>96)?14 : (
			(icount[7:0]>69)?13 : (
			(icount[7:0]>50)?12 : (
			(icount[7:0]>36)?11 : (
			(icount[7:0]>26)?10 : (
			(icount[7:0]>19)?9 : (
			(icount[7:0]>13)?8 : (
			(icount[7:0]>10)?7 : (
			(icount[7:0]>=7)?6 : (
			(icount[7:0]>=5)?5 : icount[3:0] ))))))))));

// sequential logic
always @(posedge clk) begin
	if (init) begin
		pixcnt<=(pixcnt==(512*480-1))?0:pixcnt+1;	// pixel count
		pix3<={pix3[7:0],pix};				// pixel shifting
		// update: c=x-256*delta or c=c+delta
		c<=(pixcnt[8:0]==511)? (x-{delta,8'h0}):(c+delta);
		if (pixcnt[8:0]==511) begin
			// update: d=y+240*delta or d=d-delta
			d<=(pixcnt[17:9]==479)?(y+{delta,8'h0}-{delta,4'h0}):(d-delta);
		end
		if (pixcnt[0]&pixcnt[1]) begin	// every 4 pixels...
			pix4<={pix3,pix};
			waddr<=pixcnt[17:2];
			wpend<=1;
		end
	end
	if (wpend) begin
		if (~vrd) wpend<=0;	// write pending can be delayed if video reads memory
	end
end


//////////////////////////////////
// VIDEO generator
//////////////////////////////////
wire vrd;		// video reads memory if 1
wire [15:0]va;	// video address for read

videomodule video0(.clk(clk), .d(xdi), .a(va), .vrd(vrd),
	.hsyn(hsyn), .vsyn(vsyn), .video(video)
);

//////////////////////////////////
// external RAM
//////////////////////////////////
assign xdo=pix4;			// write data
assign xa=vrd?va:waddr;		// memory address
assign xoe=~vrd;			// output enable (active low)
assign xwe=~((wpend&(~vrd))&(~clk));	// Write enable (active low)

/////////////////////////////////////////
// NES game controller and position logic
/////////////////////////////////////////

reg [2:0]gcnt=0;
reg [7:0]gbut;

assign gclk=vsyn;
assign glat=gclk&(gcnt==0);

always @(posedge gclk) begin
	gcnt<=gcnt+1;
	if (gcnt==7) begin
		if 		(gbut[0]) x<=x-{delta,4'h0};
		else if (gbut[1]) x<=x+{delta,4'h0};
		if 		(gbut[2]) y<=y+{delta,4'h0};
		else if (gbut[3]) y<=y-{delta,4'h0};
		if 		(gbut[6]) delta<={delta[11:0],delta[12]};
		else if (gbut[7]) delta<={delta[0],delta[12:1]};
	end
end
always @(negedge gclk) begin
	gbut<={gbut[6:0],~gin};
end

endmodule


